###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project

from starthinker.task.dv_editor.advertiser import advertiser_clear
from starthinker.task.dv_editor.advertiser import advertiser_load
from starthinker.task.dv_editor.audit import audit_clear
from starthinker.task.dv_editor.audit import audit_load
from starthinker.task.dv_editor.partner import partner_clear
from starthinker.task.dv_editor.partner import partner_load
from starthinker.task.dv_editor.line_item import line_item_clear

from starthinker.task.dv_targeter.campaign import campaign_clear
from starthinker.task.dv_targeter.campaign import campaign_load
from starthinker.task.dv_targeter.insertion_order import insertion_order_clear
from starthinker.task.dv_targeter.insertion_order import insertion_order_load
from starthinker.task.dv_targeter.line_item import line_item_targeting_load
from starthinker.task.dv_targeter.partner import partner_targeting_load
from starthinker.task.dv_targeter.advertiser import advertiser_targeting_load

from starthinker.task.dv_targeter.types import DV360_Targeting
from starthinker.task.dv_targeter.types import TARGETING_TYPES

from starthinker.util.bigquery import table_create
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.sheets import sheets_clear


def targeting_clear():
  targeting = DV360_Targeting()

  for targeting_meta in TARGETING_TYPES.values():

    table_create(
      project.task['auth_bigquery'],
      project.id,
      project.task['dataset'],
      'DV_Targeting_%s' % targeting_meta['name'],
      targeting.get_schema(targeting_meta['name'], targeting_meta['resource'])
    )

    sheets_clear(
      project.task['auth_sheets'],
      project.task['sheet'],
      targeting_meta['name'],
      'A2:Z'
    )


@project.from_parameters
def dv_targeter():
  print('COMMAND:', project.task['command'])

  if project.task['command'] == 'Clear':
    audit_clear()
    #patch_clear()
    partner_clear()
    targeting_clear()

  elif project.task['command'] == 'Load Partners':
    partner_clear()
    partner_load()

  elif project.task['command'] == 'Load Advertisers':
    advertiser_clear()
    advertiser_load()

  elif project.task['command'] == 'Load Line Items':
    campaign_clear()
    campaign_load()
    insertion_order_clear()
    insertion_order_load()
    line_item_clear()
    line_item_load()

  elif project.task['command'] == 'Load Targeting':
    #targeting_clear()
    #partner_targeting_load()
    #advertiser_targeting_load()
    line_item_targeting_load()

  elif project.task['command'] in ('Preview', 'Patch'):
    audit_clear()
    patch_clear()
    audit_load()

    #partner_patch(commit=project.task['command'] == 'Patch')
    #advertiser_patch(commit=project.task['command'] == 'Patch')
    #line_item_patch(commit=project.task['command'] == 'Patch')


if __name__ == '__main__':
  dv_targeter()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import argparse
import textwrap

from starthinker.util.project import project
from starthinker.util.email import send_email
from starthinker.util.email.template import EmailTemplate


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Command line to send email template via gMail.

      Email templates are JSON that assembles into both HTMl and TXT parts of an email.
      For email sample see: https://github.com/google/starthinker/blob/master/starthinker/task/newsletter/sample.json

      Example:
        - Generate an HTML page from a template, then view via browser.
          python helper.py --template sample.json > ~/Downloads/email.html

        - Send an email template via gMail.
          python helper.py --template sample.json --email_to kenjora@google.com --email_from kenjora@google.com -u $STARTHINKER_USER
  """))

  # get parameters
  parser.add_argument(
      '--template',
      help='template to use for email',
      default=None,
      required=True)
  parser.add_argument('--email_to', help='email to', default=None)
  parser.add_argument('--email_from', help='email from', default=None)

  # initialize project
  project.from_commandline(parser=parser, arguments=('-u', '-c', '-v'))

  # load template
  with open(project.args.template, 'r') as json_file:
    email = EmailTemplate(json.load(json_file))

  # send or print
  if project.args.email_to and project.args.email_from:
    print('EMAILING: ', project.args.email_to)
    send_email('user', project.args.email_to, project.args.email_from, None,
               email.get_subject(), email.get_text(), email.get_html())
  else:
    # write to STDOUT
    print(email.get_html())
    print('<pre style="width:600px;margin:0px auto;">%s</pre>' %
          email.get_text())


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.sdf import sdf_download, sdf_to_bigquery


@project.from_parameters
def sdf():
  if project.verbose:
    print('SDF')

  # Download sdf files
  sdf_zip_file = sdf_download(project.task['auth'], project.task['version'],
                              project.task['partner_id'],
                              project.task['file_types'],
                              project.task['filter_type'],
                              project.task['read']['filter_ids'])

  # Load data into BigQuery
  sdf_to_bigquery(sdf_zip_file, project.id, project.task['dataset'],
                  project.task['time_partitioned_table'],
                  project.task['create_single_day_table'],
                  project.task.get('table_suffix', ''))


if __name__ == '__main__':
  sdf()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import re
import json
import datetime

from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import Flow, InstalledAppFlow
from google.oauth2.credentials import Credentials as CredentialsUser
from google.oauth2.service_account import Credentials as CredentialsService

from starthinker.config import APPLICATION_NAME, APPLICATION_SCOPES
from starthinker.util.auth.storage import credentials_storage_get, credentials_storage_put

RE_CREDENTIALS_BUCKET = re.compile(r'[a-z0-9_\-\.]+:.+\.json')
RE_CREDENTIALS_JSON = re.compile(r'^\s*\{.*\}\s*$', re.DOTALL)


def CredentialsFlowWrapper(client, credentials_only=False, **kwargs):

  # relax scope comparison, order and default scopes are not critical
  os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = '1'

  # parse credentials from file or json
  if RE_CREDENTIALS_JSON.match(client):
    client_json = json.loads(client)
  else:
    with open(client, 'r') as json_file:
      client_json = json.load(json_file)

  if credentials_only:
    return client_json
  else:
    if 'installed' in client_json:
      flow = InstalledAppFlow.from_client_config(client_json,
                                                 APPLICATION_SCOPES, **kwargs)
    else:
      flow = Flow.from_client_config(client_json, APPLICATION_SCOPES, **kwargs)

    flow.user_agent = APPLICATION_NAME

    return flow


def CredentialsServiceWrapper(service):
  if isinstance(service, dict):
    return CredentialsService.from_service_account_info(service)
  elif RE_CREDENTIALS_JSON.match(service):
    return CredentialsService.from_service_account_info(json.loads(service))
  else:
    return CredentialsService.from_service_account_file(service)


class CredentialsUserWrapper(CredentialsUser):

  def __init__(self, user=None, client=None):
    self.user = user
    self.client = client

    super(CredentialsUserWrapper, self).__init__(None)
    self.load()

  def from_credentials(self, credentials):
    self.token = credentials.token
    self.expiry = credentials.expiry
    self._refresh_token = credentials.refresh_token
    self._id_token = credentials.token_uri
    self._token_uri = credentials.token_uri
    self._client_id = credentials.client_id
    self._client_secret = credentials.client_secret
    self._scopes = credentials.scopes

  def from_json(self, data):
    self.token = data['access_token']
    self.expiry = datetime.datetime.strptime(
        data['token_expiry'][:19],
        '%Y-%m-%dT%H:%M:%S').replace(microsecond=0)  # this is always UTC
    self._refresh_token = data['refresh_token']
    self._id_token = data['id_token']
    self._token_uri = data['token_uri']
    self._client_id = data['client_id']
    self._client_secret = data['client_secret']

  def to_json(self):
    return {
        'access_token':
            self.token,
        'token_expiry':
            self.expiry.strftime('%Y-%m-%dT%H:%M:%SZ') if self.expiry else None,
        'refresh_token':
            self._refresh_token,
        'id_token':
            self._id_token,
        'token_uri':
            self._token_uri,
        'client_id':
            self._client_id,
        'client_secret':
            self._client_secret,
        'scopes':
            self._scopes,
    }

  def load_file(self):
    if os.path.exists(self.user):
      with open(self.user, 'r') as json_file:
        self.from_json(json.load(json_file))
    elif self.client:
      self.load_flow()

  def load_storage(self):
    return self.from_json(credentials_storage_get(self.user))

  def load_flow(self):
    flow = CredentialsFlowWrapper(self.client)
    flow.run_console()

    self.from_credentials(flow.credentials)
    self.save()

  def load(self):
    if self.user is None:
      pass
    elif isinstance(self.user, dict):
      self.from_json(self.user)
    elif RE_CREDENTIALS_JSON.match(self.user):
      self.from_json(json.loads(self.user))
    elif RE_CREDENTIALS_BUCKET.match(self.user):
      self.load_storage()
    elif self.user:
      self.load_file()

  def save_json(self):
    self.user = json.dumps(self.to_json())

  def save_file(self):
    with open(self.user, 'w') as json_file:
      json_file.write(json.dumps(self.to_json()))

  def save_storage(self):
    credentials_storage_put(self.user, self.to_json())

  def save(self, destination=None):
    if destination is not None:
      self.user = destination

    if self.user is None:
      pass
    elif isinstance(self.user, dict):
      self.user = self.to_json()
    elif RE_CREDENTIALS_BUCKET.match(self.user):
      self.save_storage()
    elif RE_CREDENTIALS_JSON.match(self.user):
      self.save_json()
    elif self.user:
      self.save_file()

  def refresh(self, request=None):
    self.load()
    if not self.valid:
      if request is None:
        request = Request()
      super(CredentialsUserWrapper, self).refresh(request)
      self.expiry = self.expiry.replace(
          microsecond=0
      )  # make parsing more consistent, microseconds are not essential
      self.save()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import textwrap
import argparse

from starthinker.util.project import project
from starthinker.util.bigquery import get_schema
from starthinker.util.bigquery import rows_to_table
from starthinker.util.bigquery import table_to_schema
from starthinker.util.csv import csv_to_rows
from starthinker.util.csv import excel_to_rows


def main():
  # get parameters
  parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent("""\
    Command line to get table schema from BigQuery.

    Helps developers upload data to BigQuery and pull schemas.  These are the
    most common BigQuery tasks when developing solutions.

    Examples:
      Display table schema: `python helper.py --project [id] --dataset [name] --table [name] -s [credentials]`
      Upload csv table: `python helper.py --project [id] --dataset [name] --table [name] --csv [file] --schema [file] -s [credentials]`
      Upload excel sheet: `python helper.py --project [id] --dataset [name] --table [name] --excel_file [file] --excel_sheet [name] --schema [file] -s [credentials]`

  """))

  parser.add_argument(
    '--dataset',
    help='name of BigQuery dataset',
    default=None
  )
  parser.add_argument(
    '--table',
    help='name of BigQuery table',
    default=None
  )
  parser.add_argument(
    '--csv',
    help='CSV file path',
    default=None
  )
  parser.add_argument(
    '--schema',
    help='SCHEMA file path',
    default=None
  )
  parser.add_argument(
    '--excel_workbook',
    help='Excel file path',
    default=None
  )
  parser.add_argument(
    '--excel_sheet',
    help='Excel sheet name',
    default=None
  )

  # initialize project
  project.from_commandline(
    parser=parser,
    arguments=('-u', '-c', '-s', '-v', '-p')
  )

  auth = 'service' if project.args.service else 'user'

  schema = json.loads(project.args.schema) if project.args.schema else None

  if project.args.csv:

    with open(project.args.csv, 'r') as csv_file:
      rows = csv_to_rows(csv_file.read())

      if not schema:
        rows, schema = get_schema(rows)
        print('DETECETED SCHEMA', json.dumps(schema))
        print('Please run again with the above schema provided.')
        exit()

      rows_to_table(
        auth,
        project.id,
        project.args.dataset,
        project.args.table,
        rows,
        schema
      )

  elif project.args.excel_workbook and project.args.excel_sheet:
    with open(project.args.excel_workbook, 'r') as excel_file:
      rows = excel_to_rows(excel_file, project.args.excel_sheet)

      if not schema:
        rows, schema = get_schema(rows)
        print('DETECETED SCHEMA', json.dumps(schema))
        print('Please run again with the above schema provided.')
        exit()

      rows_to_table(
        auth,
        project.id,
        project.args.dataset,
        project.args.table,
        rows,
        schema
      )

  else:
    # print schema
    print(json.dumps(
      table_to_schema(
        auth,
        project.id,
        project.args.dataset,
        project.args.table
      ),
      indent=2
    ))

if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# http://stackoverflow.com/questions/25832631/download-attachments-from-gmail-using-gmail-api

import re
import base64
from email.encoders import encode_base64
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from io import BytesIO
from urllib.request import urlopen
from html.parser import HTMLParser
from datetime import timedelta

from googleapiclient.errors import HttpError

from starthinker.util.project import project
from starthinker.util.google_api import API_Gmail
from starthinker.util.regexp import parse_url, date_to_str
from starthinker.util.storage import parse_filename
from starthinker.util.csv import rows_to_csv


def _list_unique(seq):
  seen = set()
  seen_add = seen.add
  return [x for x in seq if not (x in seen or seen_add(x))]


def get_subject(message):
  for header in message['payload']['headers']:
    if header.get('name', '').lower() == 'subject':
      return header['value']
  return ''


def get_email_links(auth, message, link_regexp, download=False):
  links = []
  html_parser = HTMLParser()
  link_filter = re.compile(r'%s' % link_regexp) if link_regexp else None

  try:
    for part in message['payload'].get('parts', []) or [message['payload']]:
      if 'data' in part['body']:
        data = part['body']['data']
        content = base64.urlsafe_b64decode(data).decode('utf-8')
        # plain text may be different than html
        if part['mimeType'] == 'text/plain':
          links.extend(parse_url(content))
        # html needs to decode links
        elif part['mimeType'] == 'text/html':
          links.extend(
              map(lambda link: html_parser.unescape(link), parse_url(content)))

  except HttpError as error:
    print('EMAIL LINK ERROR: %s' % error)

  # remove duplicates
  links = _list_unique(links)

  # filter links
  if link_filter:
    links = [link for link in links if link_filter.match(link)]

  # for downloads convert links into files and data
  for link in links:
    if download:
      try:
        yield parse_filename(link, url=True), BytesIO(urlopen(link).read())
      except:
        'ERROR: Unable To Download', link
    else:
      yield link


def get_email_attachments(auth, message, attachment_regexp):

  file_filter = re.compile(r'%s' %
                           attachment_regexp) if attachment_regexp else None

  try:
    for part in message['payload'].get('parts', []):

      if part['filename']:

        # filter regexp
        if not file_filter or file_filter.match(part['filename']):

          if 'data' in part['body']:
            data = part['body']['data']

          else:
            att_id = part['body']['attachmentId']
            att = API_Gmail(auth).users().messages().attachments().get(
                userId='me', messageId=message['id'], id=att_id).execute()
            data = att['data']

          file_data = BytesIO(base64.urlsafe_b64decode(data.encode('UTF-8')))
          yield part['filename'], file_data

  except HttpError as e:
    print('EMAIL ATTACHMENT ERROR:', str(e))


def get_email_messages(auth,
                       email_from,
                       email_to,
                       subject_regexp=None,
                       date_min=None,
                       date_max=None):
  if project.verbose:
    print('GETTING EMAILS')

  query = 'from:%s AND to:%s' % (email_from, email_to)
  if date_min:
    query += ' AND after:%s' % date_to_str(date_min)
  if date_max:
    query += ' AND before:%s' % date_to_str(
        date_max + timedelta(days=1))  # make it inclusive
  if project.verbose:
    print('EMAIL SEARCH:', query)

  messages = API_Gmail(
      auth, iterate=True).users().messages().list(
          userId='me', q=query).execute()

  subject_filter = re.compile(r'%s' %
                              subject_regexp) if subject_regexp else None
  for message in messages:
    message = API_Gmail(auth).users().messages().get(
        userId='me', id=message['id']).execute()
    if subject_filter is None or subject_filter.match(get_subject(message)):
      yield message


def send_email(auth,
               email_to,
               email_from,
               email_cc,
               subject,
               text,
               html=None,
               attachment_filename=None,
               attachment_rows=None):
  if project.verbose:
    print('SENDING EMAIL', email_to)

  message = MIMEMultipart('alternative')
  message.set_charset('utf8')

  message['to'] = email_to
  message['cc'] = email_cc
  message['from'] = email_from
  message['subject'] = subject
  message.attach(MIMEText(text, 'plain', 'UTF-8'))

  if html:
    message.attach(MIMEText(html, 'html', 'UTF-8'))

  if attachment_filename and attachment_rows:
    attachment = MIMEBase('text', 'csv')
    attachment.set_payload(rows_to_csv(attachment_rows).read())
    attachment.add_header(
        'Content-Disposition', 'attachment', filename=attachment_filename)
    encode_base64(attachment)
    message.attach(attachment)

  #API_Gmail(auth).users().messages().send(userId='me', body={'raw': base64.urlsafe_b64encode(message.as_string())}).execute()
  API_Gmail(auth).users().messages().send(
      userId='me',
      body={
          'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()
      }).execute()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.dbm import sdf_read
from starthinker.util.bigquery import query_to_table
from starthinker.util.sdf import sdf_schema
from starthinker.util.sdf.schema.Lookup import SDF_Field_Lookup
from starthinker.util.csv import column_header_sanitize
from starthinker.util.data import put_rows, get_rows


@project.from_parameters
def sdf_legacy():
  if project.verbose:
    print('SDF')

  # Read Filter Ids
  filter_ids = list(
      get_rows(project.task['auth'], project.task['read']['filter_ids']))

  # Loop through requested file types
  for file_type in project.task['file_types']:
    disposition = 'WRITE_TRUNCATE'

    # if daily then create a seperate table for each day else accumulate all in one table
    table = ('SDF_%s_%s' % (file_type, str(project.date).replace('-', '_'))
            ) if project.task['daily'] else ('SDF_%s' % file_type)

    # do one filter id at a time to avoid response too large error ( product knows )
    for filter_id in filter_ids:

      if project.verbose:
        print('SDF DOWNLOAD', project.task['filter_type'], file_type, filter_id)
      rows = sdf_read(project.task['auth'], [file_type],
                      project.task['filter_type'], [filter_id],
                      project.task.get('version', '5'))

      if rows:
        schema = sdf_schema(next(rows))
        if 'bigquery' in project.task['out']:
          project.task['out']['bigquery']['schema'] = schema
          project.task['out']['bigquery']['skip_rows'] = 0
          project.task['out']['bigquery']['table'] = table

        put_rows(project.task['auth'], project.task['out'], rows)

      else:
        if project.verbose:
          print('NO DATA')

    disposition = 'WRITE_APPEND'

    if project.task['daily']:

      if project.verbose:
        print('SDF COMBINE DAYS', file_type)
      # Load data into BigQuery
      query_to_table(
          project.task['auth'],
          project.id,
          project.task['out']['bigquery']['dataset'],
          'SDF_%s' % file_type,
          "SELECT PARSE_DATE('%%Y_%%m_%%d',_TABLE_SUFFIX) as SDF_Day, * FROM `%s.%s.SDF_%s_*`"
          % (project.id, project.task['out']['bigquery']['dataset'], file_type),
          disposition='WRITE_TRUNCATE',
          legacy=False)


if __name__ == '__main__':
  sdf_legacy()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Test Script

Used by tests.

  - This should be called by the tests scripts only.
  - When run will generate a say hello log.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_read': 'user',  # Credentials used for reading data.
}

RECIPE = {
  'setup': {
    'day': [
    ],
    'hour': [
    ]
  },
  'tasks': [
    {
      'hello': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'hour': [
        ],
        'say': 'Hello Manual',
        'sleep': 0
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('manual', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.campaign import campaign_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def frequency_cap_clear():
  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Frequency Caps",
               "A2:Z")


def frequency_cap_load():

  # write frequency_caps to sheet
  rows = get_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "query":
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         NULL,
         NULL,
         C.frequencyCap.unlimited,
         C.frequencyCap.unlimited,
         C.frequencyCap.timeUnit,
         C.frequencyCap.timeUnit,
         C.frequencyCap.timeUnitCount,
         C.frequencyCap.timeUnitCount,
         C.frequencyCap.maxImpressions,
         C.frequencyCap.maxImpressions
       FROM `{dataset}.DV_Campaigns` AS C
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON C.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         NULL,
         I.frequencyCap.unlimited,
         I.frequencyCap.unlimited,
         I.frequencyCap.timeUnit,
         I.frequencyCap.timeUnit,
         I.frequencyCap.timeUnitCount,
         I.frequencyCap.timeUnitCount,
         I.frequencyCap.maxImpressions,
         I.frequencyCap.maxImpressions
       FROM `{dataset}.DV_InsertionOrders` AS I
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON I.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON I.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         CONCAT(L.displayName, ' - ', L.lineItemId),
         L.frequencyCap.unlimited,
         L.frequencyCap.unlimited,
         L.frequencyCap.timeUnit,
         L.frequencyCap.timeUnit,
         L.frequencyCap.timeUnitCount,
         L.frequencyCap.timeUnitCount,
         L.frequencyCap.maxImpressions,
         L.frequencyCap.maxImpressions
       FROM `{dataset}.DV_LineItems` AS L
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON L.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
       ON L.insertionOrderId=I.insertionOrderId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON L.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       """.format(**project.task),
              "legacy":
                  False
          }
      })

  put_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Frequency Caps",
              "range": "A2"
          }
      }, rows)


def frequency_cap_audit():
  rows = get_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Frequency Caps",
              "range": "A2:Z"
          }
      })

  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset": project.task["dataset"],
              "table": "SHEET_FrequencyCaps",
              "schema": [
                  { "name": "Partner", "type": "STRING" },
                  { "name": "Advertiser", "type": "STRING" },
                  { "name": "Campaign", "type": "STRING" },
                  { "name": "Insertion_Order", "type": "STRING" },
                  { "name": "Line_Item", "type": "STRING" },
                  { "name": "Unlimited", "type": "BOOLEAN" },
                  { "name": "Unlimited_Edit", "type": "BOOLEAN" },
                  { "name": "Time_Unit", "type": "STRING" },
                  { "name": "Time_Unit_Edit", "type": "STRING" },
                  { "name": "Time_Count", "type": "INTEGER" },
                  { "name": "Time_Count_Edit", "type": "INTEGER" },
                  { "name": "Max_impressions", "type": "INTEGER" },
                  { "name": "Max_impressions_Edit", "type": "INTEGER" },
              ],
              "format": "CSV"
          }
      }, rows)

  query_to_view(
      project.task["auth_bigquery"],
      project.id,
      project.task["dataset"],
      "AUDIT_FrequencyCaps",
      """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Frequency Caps' AS Operation,
            CASE
              WHEN Unlimited_Edit IS NULL THEN 'Missing Unlimited.'
              WHEN Time_Unit_Edit IS NULL THEN 'Missing Time Unit.'
              WHEN Time_Count_Edit IS NULL THEN 'Missing Time Count.'
              WHEN Max_Impressions_Edit IS NULL THEN 'Missing Max Impressions.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Line_Item, Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_FrequencyCaps`
        )
        WHERE
          Error IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      ;
    """.format(**project.task),
      legacy=False)

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_FrequencyCaps",
    """SELECT *
      FROM `{dataset}.SHEET_FrequencyCaps`
      WHERE Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_FrequencyCaps` WHERE Severity='ERROR')
      AND Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_FrequencyCaps` WHERE Severity='ERROR')
      AND Campaign NOT IN (SELECT Id FROM `{dataset}.AUDIT_FrequencyCaps` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def frequency_cap_patch(commit=False):
  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_FrequencyCaps",
    }},
    as_object=True
  )

  for row in rows:

    frequency_cap = {}

    if row['Unlimited'] != row['Unlimited_Edit']:
      frequency_cap.setdefault("frequencyCap", {})
      frequency_cap["frequencyCap"]["unlimited"] = row['Unlimited_Edit']
    if row['Time_Unit'] != row['Time_Unit_Edit']:
      frequency_cap.setdefault("frequencyCap", {})
      frequency_cap["frequencyCap"]["timeUnit"] = row['Time_Unit_Edit']
    if row['Time_Count'] != row['Time_Count_Edit']:
      frequency_cap.setdefault("frequencyCap", {})
      frequency_cap["frequencyCap"]["timeUnitCount"] = row['Time_Count_Edit']
    if row['Max_impressions'] != row['Max_impressions_Edit']:
      frequency_cap.setdefault("frequencyCap", {})
      frequency_cap["frequencyCap"]["maxImpressions"] = row['Max_impressions_Edit']

    if frequency_cap:
      patch = {
          "operation": "Frequency Caps",
          "action": "PATCH",
          "partner": row['Partner'],
          "advertiser": row['Advertiser'],
          "parameters": {
              "advertiserId": lookup_id(row['Advertiser']),
              "body": frequency_cap
          }
      }

      if row['Line_Item']:
        patch["line_item"] = row['Line_Item']
        patch["parameters"]["lineItemId"] = lookup_id(row['Line_Item'])
      elif row['Insertion_Order']:
        patch["insertion_order"] = row['Insertion_Order']
        patch["parameters"]["insertionOrderId"] = lookup_id(row['Insertion_Order'])
      else:
        patch["campaign"] = row['Campaign']
        patch["parameters"]["campaignId"] = lookup_id(row['Campaign'])

      patches.append(patch)

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
    line_item_commit(patches)
    campaign_commit(patches)
  else:
    patch_preview(patches)



import pprint
import json
import pytz
import traceback
import dateutil
import requests
import re

from datetime import datetime
from tzlocal import get_localzone
from starthinker.util.project import project
from starthinker.util.google_api import API_DCM
from starthinker.util.email import send_email
from starthinker.task.traffic.feed import Feed, FieldMap

DATE_FORMAT = '%m/%d/%Y %H:%M:%S'


def is_live(url):
  try:
    request_headers = {
        'user-agent':
            'Mozilla/5.0 (X11; CrOS x86_64 11151.113.1) AppleWebKit/537.36 '
            '(KHTML, like Gecko) Chrome/71.0.3578.127 Safari/537.36',
        'Accept':
            '*/*'
    }

    response = requests.get(url, headers=request_headers, timeout=5)

    return_code = response.status_code
    response.close()

    return return_code >= 200 and return_code <= 299
  except Exception as ex:
    print(str(ex))

    return False


def write_status(status, message):
  status_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'lp_dawn_status',
      parse=False)

  if len(status_feed.feed) == 1:
    status_feed_item = status_feed.feed[0]

    now = datetime.now(get_localzone())
    now = now.astimezone(pytz.timezone('America/Chicago'))

    status_feed_item['Last Run Time'] = now.strftime('%Y-%m-%dT%H:%M:%S.000%z')
    status_feed_item['Last Run Status'] = status
    status_feed_item['Message'] = message

    status_feed.update()


@project.from_parameters
def lp_dawn():
  if project.verbose:
    print('Landing Page Dawn')

  # Load feed
  feed = None
  status = 'NORMAL'
  message = ''

  try:
    feed = Feed(
        project.task['auth'], project.task['sheet_id'], 'lp_dawn', parse=False)
  except:
    pass

  # For each item in the feed
  try:
    for feed_item in feed.feed:
      try:
        # Verify the status of the status of the item to determine if it needs
        # processing
        if feed_item[FieldMap.LP_DAWN_STATUS] in ['', 'PENDING', 'LP_NOT_LIVE']:

          # Verify if the date is in the past
          if feed_item[FieldMap.LP_DAWN_TIME_FOR_UPDATE]:
            date = dateutil.parser.parse(
                '%sT%s' % (feed_item[FieldMap.LP_DAWN_DATE_FOR_UPDATE],
                           feed_item[FieldMap.LP_DAWN_TIME_FOR_UPDATE]))
          else:
            date = dateutil.parser.parse(
                '%s' % (feed_item[FieldMap.LP_DAWN_DATE_FOR_UPDATE]))

          date = pytz.timezone('America/Chicago').localize(date)
          now = datetime.now(get_localzone())

          if date < now:

            # Verify if the landing page is live
            if feed_item.get('Force', 'false').lower() == 'true' or is_live(
                feed_item[FieldMap.LP_DAWN_URL]):

              # Update Landing Page
              landing_page = API_DCM(project.task['auth']).advertiserLandingPages().get(
                  profileId=project.task['dcm_profile_id'],
                  id=feed_item[FieldMap.LP_DAWN_LP_ID]).execute()
              landing_page['url'] = feed_item[FieldMap.LP_DAWN_URL]
              API_DCM(project.task['auth']).advertiserLandingPages().update(
                  profileId=project.task['dcm_profile_id'],
                  body=landing_page).execute()

              # Update status in the feed
              feed_item[FieldMap.LP_DAWN_STATUS] = 'UPDATED'
              feed_item[FieldMap.LP_DAWN_UPDATE_TIME] = now.strftime(
                  '%Y-%m-%dT%H:%M:%S.000%z')
              feed_item[FieldMap.LP_DAWN_MESSAGE] = ''

            else:
              feed_item[FieldMap.LP_DAWN_STATUS] = 'PENDING'
              feed_item[
                  FieldMap.LP_DAWN_MESSAGE] = '%s is not live' % feed_item[
                      FieldMap.LP_DAWN_URL]

              if status == 'NORMAL':
                status = 'WARNING'
                message = 'One or more landing pages are not live'
          else:
            feed_item[FieldMap.LP_DAWN_STATUS] = 'PENDING'
            feed_item[
                FieldMap.LP_DAWN_MESSAGE] = 'Date / Time for update not reached'
      except Exception as e:
        traceback.print_exc()
        msg = str(e)

        status = 'ERROR'
        message = msg

        try:
          print('sending email')
          print(feed_item)
          print(feed_item[FieldMap.LP_DAWN_EMAIL])
          send_email(
              'user', feed_item[FieldMap.LP_DAWN_EMAIL], 'mauriciod@google.com',
              '', 'Landing Page Dawn Error Notification',
              '%s: Landing Page ID: %s' %
              (msg, feed_item[FieldMap.LP_DAWN_LP_ID]))
        except:
          pass

        match = re.search(r'"(.*)"', msg)

        if match:
          feed_item[FieldMap.LP_DAWN_MESSAGE] = 'ERROR: %s' % match.group(1)
        else:
          feed_item[FieldMap.LP_DAWN_MESSAGE] = 'ERROR: %s' % msg

        feed_item[FieldMap.LP_DAWN_STATUS] = 'ERROR'

  except Exception as ex:
    status = 'ERROR'
    message = str(e)
    send_email('user', 'mauriciod@google.com', 'mauriciod@google.com', '',
               'Landing Page Dawn Error Notification', 'Unhandled error!')

  finally:
    feed.update()
    try:
      write_status(status, message)
    except:
      pass

  # handle errors, do not fail silently


if __name__ == '__main__':
  lp_dawn()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Standard_Report_Filters_Schema = [{
    'name': 'Activity',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Attributed_Event_Connection_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Attributed_Event_Environment',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_External_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'City',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Companion_Creative',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Companion_Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Connection_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Content_Category',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Content_Classifier',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_4',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_5',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_6',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_7',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_8',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_9',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_10',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_11',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Field_12',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Digital_Content_Label',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Designated_Market_Area_Dma',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Domain',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_4',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_5',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Field_Value_6',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Environment',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Attributed_Event_Platform_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Configuration',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Hour',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Mobile_Carrier',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Property',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Query',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Total_Booked_Units',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Agency',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Landing_Page_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Payment_Source',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Compatibility',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Cost_Structure',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_External_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Tag_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Total_Booked_Units',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Platform_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rendering_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Dcm',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Site_Directory',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'State_Region',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'User_List',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'User_List_Description',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'User_List_Membership_Life_Span',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Prominence_Score',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.sheets import sheets_clear

from starthinker.tasks.dv_targeting.types import TARGETING_SCHEMA
from starthinker.tasks.dv_targeting.types import TARGETING_TYPES

def clear():
  for targeting_type in TARGETING_TYPES:
    table_create(
      project.task['auth_bigquery'],
      project.id,
      project.task['dataset'],
      'DV_Targeting_%s' % targeting_type,
      TARGETING_SCHEMA[targeting_type]
    )

    sheets_clear(
      project.task['auth_sheets'],
      project.task['sheet'],
      targeting_type,
      'A2:Z'
    )



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import pysftp
import gzip
import datetime
import uuid
import os
import csv

from starthinker.util.project import project
from starthinker.util.bigquery import io_to_table, make_schema


def clean_csv(input_file,
              output_file_name,
              num_fields,
              header=True,
              append_to_lines=''):
  if os.path.isfile(output_file_name):
    os.remove(output_file_name)

  output_file = open(output_file_name, 'w')

  for line in input_file:
    if header:
      header = False
    elif len(line) > 1:
      line = line.strip()
      while line.count(',') < (num_fields - 1):
        line += ','
      output_file.write(line + append_to_lines + '\n')
    else:
      break

  output_file.close()


# loop all parameters and replace with values, for lists turn them into strings
def query_parameters(query, parameters):
  while '[PARAMETER]' in query:
    parameter = parameters.pop(0)
    if isinstance(parameter, list) or isinstance(parameter, tuple):
      parameter = ', '.join([str(p) for p in parameter])
    query = query.replace('[PARAMETER]', parameter, 1)
  if project.verbose:
    print('QUERY:', query)
  return query


@project.from_parameters
def sftp():

  cnopts = pysftp.CnOpts()
  cnopts.hostkeys = None
  sftp_configs = project.task['from']['sftp']['connection']
  sftp_configs['cnopts'] = cnopts
  sftp = pysftp.Connection(**sftp_configs)

  file_name = (
      datetime.datetime.now() +
      datetime.timedelta(project.task['from']['sftp'].get('day', 0))).strftime(
          project.task['from']['sftp']['file'])
  input_file_name = '/tmp/%s.csv' % str(uuid.uuid1())
  sftp.get(file_name, localpath=input_file_name)

  compression = project.task['from']['sftp'].get('compression', None)

  if 'table' in project.task['to']:
    input_file = None
    if compression == 'gzip':
      input_file = gzip.open(input_file_name, 'rb')
      uncompressed_file = '/tmp/%s.csv' % str(uuid.uuid1())

      out = open(uncompressed_file, 'wb')
      for line in input_file:
        if len(line) > 1:
          out.write(line)
      out.close()
      input_file.close()

      os.remove(input_file_name)
      input_file_name = uncompressed_file

    input_file = open(input_file_name, 'rb')

    reader = csv.reader(input_file)
    header = next(reader)
    input_file.seek(0)
    schema = make_schema(header)
    output_file_name = '/tmp/%s.csv' % str(uuid.uuid1())
    clean_csv(input_file, output_file_name, len(header), header=True)
    input_file.close()

    output_file = open(output_file_name, 'rb')
    io_to_table(
        project.task['auth'],
        project.id,
        project.task['to'].get('dataset'),
        project.task['to'].get('table'),
        output_file,
        'CSV',
        schema,
        skip_rows=0,
        disposition=project.task['to'].get('write_disposition',
                                           'WRITE_TRUNCATE'))
    output_file.close()

    os.remove(input_file_name)

  os.remove(output_file_name)


if __name__ == '__main__':
  sftp()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

DV360 Feature Adoption Dashboard

Tracks revenue spent on various DV360 features and compares to performance metrics.

  - IMPORTANT: You must add Advertiser IDS, we're still working on a partner level report.
  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->DV360_Feature_Analysis</b> dataset to be created.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/c/u/0/datasources/1IPh_zqjFeZTemq6iTUvc3JvZqD0Frneo' target='_blank'>Sample DV360 Feature Adoption Analysis</a>.
  - Click Edit Connection, find above dataset, and choose <b>Sample DV360 Feature Adoption Analysis</b>.
  - Copy <a href='https://datastudio.google.com/c/u/0/datasources/1vQMz7O05gaXRX_8ZF7kKXmfaBLjVv07H' target='_blank'>Sample DV360 Feature Adoption Environment</a>.
  - Click Edit Connection, find above dataset, and choose <b>Sample DV360 Feature Adoption Environment</b>.  This report takes a long time to run, on your first setup you may have to wait a few hours for this step to complete.
  - Copy <a href='https://datastudio.google.com/open/1uZ4WtxiRsBu54gxFdYAzPfW2yGRZlWn0?usp=sharing' target='_blank'>Sample DV360 Feature Adoption Report</a>.
  - When prompted choose the new data source you just created.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be created in BigQuery.
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_name': '',  # Name of report in DV360, should be unique.
  'partners': [],  # DV360 partner id.
  'advertisers': [],  # Comma delimited list of DV360 advertiser ids.
  'recipe_project': '',  # Google Cloud Project Id.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      2,
      4,
      6
    ]
  },
  'tasks': [
    {
      'dataset': {
        'hour': [
          1
        ],
        'auth': 'service',
        'description': 'Create a dataset for bigquery tables.',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Feature Adoption Spend For ',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_GENERAL',
              'groupBys': [
                'FILTER_ADVERTISER',
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER_CURRENCY',
                'FILTER_INSERTION_ORDER',
                'FILTER_INSERTION_ORDER_NAME',
                'FILTER_LINE_ITEM',
                'FILTER_LINE_ITEM_NAME',
                'FILTER_DATE'
              ],
              'metrics': [
                'METRIC_IMPRESSIONS',
                'METRIC_BILLABLE_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_CTR',
                'METRIC_TOTAL_CONVERSIONS',
                'METRIC_LAST_CLICKS',
                'METRIC_LAST_IMPRESSIONS',
                'METRIC_REVENUE_ADVERTISER',
                'METRIC_MEDIA_COST_ADVERTISER',
                'METRIC_RICH_MEDIA_VIDEO_COMPLETIONS',
                'METRIC_RICH_MEDIA_VIDEO_PLAYS',
                'METRIC_CM_POST_CLICK_REVENUE',
                'METRIC_CM_POST_VIEW_REVENUE'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Feature Adoption Environment For ',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_GENERAL',
              'groupBys': [
                'FILTER_ADVERTISER',
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER_CURRENCY',
                'FILTER_INSERTION_ORDER',
                'FILTER_INSERTION_ORDER_NAME',
                'FILTER_LINE_ITEM',
                'FILTER_LINE_ITEM_NAME',
                'FILTER_DEVICE_TYPE',
                'FILTER_PAGE_LAYOUT',
                'FILTER_DATE'
              ],
              'metrics': [
                'METRIC_IMPRESSIONS',
                'METRIC_BILLABLE_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_TOTAL_CONVERSIONS',
                'METRIC_LAST_CLICKS',
                'METRIC_LAST_IMPRESSIONS',
                'METRIC_REVENUE_ADVERTISER',
                'METRIC_MEDIA_COST_ADVERTISER'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Feature Adoption Spend For ',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_Feature_Spend',
            'is_incremental_load': True,
            'header': True,
            'schema': [
              {
                'name': 'Advertiser_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Advertiser',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_Currency',
                'type': 'STRING'
              },
              {
                'name': 'Insertion_Order_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Insertion_Order',
                'type': 'STRING'
              },
              {
                'name': 'Line_Item_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Line_Item',
                'type': 'STRING'
              },
              {
                'name': 'Report_Day',
                'type': 'DATE'
              },
              {
                'name': 'Impressions',
                'type': 'INTEGER'
              },
              {
                'name': 'Billable_Impressions',
                'type': 'FLOAT'
              },
              {
                'name': 'Clicks',
                'type': 'INTEGER'
              },
              {
                'name': 'Click_Rate_Ctr',
                'type': 'STRING'
              },
              {
                'name': 'Total_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Post_Click_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Post_View_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Revenue_Adv_Currency',
                'type': 'FLOAT'
              },
              {
                'name': 'Media_Cost_Advertiser_Currency',
                'type': 'FLOAT'
              },
              {
                'name': 'Complete_Views_Video',
                'type': 'INTEGER'
              },
              {
                'name': 'Starts_Video',
                'type': 'INTEGER'
              },
              {
                'name': 'Cm_Post_Click_Revenue',
                'type': 'FLOAT'
              },
              {
                'name': 'Cm_Post_View_Revenue',
                'type': 'FLOAT'
              }
            ]
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Feature Adoption Environment For ',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_Feature_Environment',
            'is_incremental_load': True,
            'header': True,
            'schema': [
              {
                'name': 'Advertiser_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Advertiser',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_Currency',
                'type': 'STRING'
              },
              {
                'name': 'Insertion_Order_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Insertion_Order',
                'type': 'STRING'
              },
              {
                'name': 'Line_Item_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Line_Item',
                'type': 'STRING'
              },
              {
                'name': 'Device_Type',
                'type': 'STRING'
              },
              {
                'name': 'Environment',
                'type': 'STRING'
              },
              {
                'name': 'Report_Day',
                'type': 'DATE'
              },
              {
                'name': 'Impressions',
                'type': 'INTEGER'
              },
              {
                'name': 'Billable_Impressions',
                'type': 'FLOAT'
              },
              {
                'name': 'Clicks',
                'type': 'INTEGER'
              },
              {
                'name': 'Total_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Post_Click_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Post_View_Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Revenue_Adv_Currency',
                'type': 'FLOAT'
              },
              {
                'name': 'Media_Cost_Advertiser_Currency',
                'type': 'FLOAT'
              }
            ]
          }
        }
      }
    },
    {
      'sdf_legacy': {
        'hour': [
          4
        ],
        'auth': 'user',
        'version': '5',
        'file_types': [
          'LINE_ITEM'
        ],
        'filter_type': 'ADVERTISER_ID',
        'read': {
          'filter_ids': {
            'values': {
              'field': {
                'name': 'advertisers',
                'kind': 'integer_list',
                'order': 6,
                'default': [
                ],
                'description': 'Comma delimited list of DV360 advertiser ids.'
              }
            }
          }
        },
        'daily': True,
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          }
        }
      }
    },
    {
      'sdf_legacy': {
        'hour': [
          4
        ],
        'auth': 'user',
        'version': '5',
        'file_types': [
          'INSERTION_ORDER'
        ],
        'filter_type': 'ADVERTISER_ID',
        'read': {
          'filter_ids': {
            'values': {
              'field': {
                'name': 'advertisers',
                'kind': 'integer_list',
                'order': 6,
                'default': [
                ],
                'description': 'Comma delimited list of DV360 advertiser ids.'
              }
            }
          }
        },
        'daily': True,
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          }
        }
      }
    },
    {
      'bigquery': {
        'hour': [
          6
        ],
        'auth': 'service',
        'from': {
          'query': " SELECT   Report_Day,   Advertiser,   Advertiser_Id,   Advertiser_Currency,   Insertion_Order AS Insertion_Order_Name,   Insertion_Order_Id AS Insertion_Order_ID,   CONCAT(Insertion_Order,' (',CAST(Insertion_Order_ID AS STRING),')') AS Insertion_Order,   Line_Item AS LI_Name,   b.Line_Item_Id AS LI_ID,   CONCAT(Line_Item,' (',CAST(b.Line_Item_Id AS STRING),')') AS Line_Item,   IF(MAX(SDF_Day) OVER () = SDF_Day, 'True', 'False') AS Max_Date,   a.*,   IF(Bid_Strategy_Type = 'Fixed', 'Fixed', CONCAT(Bid_Strategy_Type, ' ', Bid_Strategy_Unit)) AS Bid_Strategy,   IF(Bid_Strategy_Type = 'Fixed', 'Fixed', 'Autobidding') AS Autobidding_Overall,   CONCAT(Pacing, ' ', Pacing_Rate) AS Pacing_Overall,   IF(Affinity_In_Market_Targeting_Include IS NULL, 'False', 'True') AS Affinity_In_Market_TF,   IF(Affinity_In_Market_Targeting_Exclude IS NULL, 'False', 'True') AS Affinity_In_Market_Exclude_TF,   IF(Custom_List_Targeting IS NULL, 'False', 'True') AS Custom_List_TF,   IF(Audience_Targeting_Include IS NULL     AND Affinity_In_Market_Targeting_Include IS NULL     AND Custom_List_Targeting IS NULL     AND Audience_Targeting_Similar_Audiences = 'False' , 'False', 'True') AS Audience_Targeting_TF,   IF(Keyword_Targeting_Exclude IS NULL     AND Keyword_List_Targeting_Exclude IS NULL, 'False', 'True') AS Keyword_Exclusions_TF,   IF(Position_Targeting_On_Screen IS NULL     AND Position_Targeting_Display_Position_In_Content IS NULL     AND Position_Targeting_Video_Position_In_Content IS NULL     AND Position_Targeting_Audio_Position_In_Content IS NULL, 'False', 'True') AS Position_Targeting_TF,   IF((Geography_Targeting_Include IS NULL       AND Geography_Regional_Location_List_Targeting_Include IS NULL), 'False', 'True') AS Geography_Targeting_TF,   IF(Language_Targeting_Include IS NULL, 'False', 'True') AS Language_Targeting_TF,   IF(Site_Targeting_Exclude IS NULL     AND Site_Targeting_Include IS NULL, 'False', 'True') AS Site_List_TF,   Impressions AS Impressions,   Billable_Impressions AS Billable_Impressions,   Clicks AS Clicks,   Total_Conversions AS Total_Conversions,   Post_Click_Conversions AS Post_Click_Conversions,   Post_View_Conversions AS Post_View_Conversions,   Revenue_Adv_Currency AS Rev_Adv_Currency,   Media_Cost_Advertiser_Currency AS Media_Cost_Adv_Currency,   Cm_Post_Click_Revenue AS CM_Post_Click_Revenue,   Cm_Post_View_Revenue AS CM_Post_View_Revenue,   (Cm_Post_Click_Revenue+Cm_Post_View_Revenue) AS CM_Total_Revenue,   IF (Bid_Strategy_Type = 'Fixed', 0, Revenue_Adv_Currency) AS Autobid_Rev,   IF (Type = 'Display', Revenue_Adv_Currency, 0) AS Display_Rev,   IF (Type = 'Video', Revenue_Adv_Currency, 0) AS Video_Rev,   IF (Type = 'TrueView', Revenue_Adv_Currency, 0) AS TrueView_Rev,   IF (Type = 'Audio', Revenue_Adv_Currency, 0) AS Audio_Rev,   IF ((STRPOS(Environment_Targeting, 'Desktop')>0), Revenue_Adv_Currency, 0) AS Desktop_Rev   FROM `[PARAMETER].[PARAMETER].SDF_LINE_ITEM` AS a   LEFT JOIN `[PARAMETER].[PARAMETER].DV360_Feature_Spend` AS b     ON DATE_SUB(a.SDF_Day, INTERVAL 1 DAY) = b.Report_Day     AND a.Line_Item_Id = b.Line_Item_Id   WHERE b.Report_Day IS NOT NULL ",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          ],
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be created in BigQuery.'
            }
          },
          'view': 'DV360_Feature_Adoption_Analysis'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('dv360_feature_adoption', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Main entry point of Bulkdozer."""

import traceback

from starthinker.util.dcm import get_profile_for_api
from starthinker.util.project import project
from starthinker.util.sheets import sheets_get

from starthinker.task.traffic.feed import Feed
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.ad import AdDAO
from starthinker.task.traffic.creative_assets import CreativeAssetDAO
from starthinker.task.traffic.creative_association import CreativeAssociationDAO
from starthinker.task.traffic.creative import CreativeDAO
from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.dynamic_targeting_key import DynamicTargetingKeyDAO
from starthinker.task.traffic.event_tag import EventTagDAO
from starthinker.task.traffic.landing_page import LandingPageDAO
from starthinker.task.traffic.placement import PlacementDAO
from starthinker.task.traffic.placement_group import PlacementGroupDAO
from starthinker.task.traffic.video_format import VideoFormatDAO
from starthinker.task.traffic.store import store
from starthinker.task.traffic.logger import logger

video_format_dao = None
landing_page_dao = None
placement_group_dao = None
campaign_dao = None
creative_association_dao = None
creative_dao = None
placement_dao = None
creative_asset_dao = None
ad_dao = None
event_tag_dao = None
dynamic_targeting_key_dao = None
spreadsheet = None

clean_run = True


def process_feed(feed_name, dao, print_field, msg='Processing'):
  """Processes a feed that represents a specific entity in the Bulkdozer feed.

  Args:
    feed_name: Name of the feed to process, refer to feed.py for the supported
      feed names.
    dao: The data access object to be used to interact with the CM API and
      update, must match the entity being updated in CM, in the sense that the
      required fields to fetch, create, and update the entity in CM must be
      included in the feed.
    print_field: Field that identifies the item, used to print status messages
      to the Log tab of the Bulkdozer feed.
    msg: Prefix message to use when writing to the Log tab of the Bulkdozer
      feed, for instance we display Processing Campaign for campaign, and
      Uploading Asset for assets.
  """
  feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      feed_name,
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  execute_feed(feed, dao, print_field, msg)


def execute_feed(feed, dao, print_field, msg='Processing'):
  """Executes a specific feed.

  Args:
    feed: Feed object representing the Bulkdozer feed to process.
    dao: The data access object to be used to interact with the CM API and
      update, must match the entity being updated in CM, in the sense that the
      required fields to fetch, create, and update the entity in CM must be
      included in the feed.
    print_field: Field that identifies the item, used to print status messages
      to the Log tab of the Bulkdozer feed.
    msg: Prefix message to use when writing to the Log tab of the Bulkdozer
      feed, for instance we display Processing Campaign for campaign, and
      Uploading Asset for assets.
  """
  global clean_run

  try:
    dao.pre_fetch(feed.feed)

    for feed_item in feed.feed:
      try:
        value = feed_item[print_field]
        print('%s %s' % (msg, value))
        logger.log('%s %s' % (msg, value))
        dao.process(feed_item)
      except Exception as error:
        clean_run = False
        stack = traceback.format_exc()
        print(stack)
        logger.log(str(error))

  finally:
    feed.update()


def setup():
  """Sets up Bulkdozer configuration and required object to execute the job."""

  if not 'dcm_profile_id' in project.task and 'account_id' in project.task:
    project.task['is_admin'], project.task[
        'dcm_profile_id'] = get_profile_for_api(project.task['auth'],
                                                project.task['account_id'])

  if not 'is_admin' in project.task:
    project.task['is_admin'] = False

  if 'sheet_url' in project.task and not 'sheet_id' in project.task:
    project.task['sheet_id'] = project.task['sheet_url']

  # Setting up required objects and parsing parameters
  logger.auth = project.task['auth']
  logger.trix_id = project.task.get('logger', {}).get('sheet_id',
                                                      project.task['sheet_id'])
  logger.buffered = True


def init_daos():
  global video_format_dao
  global landing_page_dao
  global placement_group_dao
  global campaign_dao
  global creative_association_dao
  global creative_dao
  global placement_dao
  global creative_asset_dao
  global ad_dao
  global event_tag_dao
  global dynamic_targeting_key_dao
  global spreadsheet

  spreadsheet = sheets_get(project.task['auth'], project.task['sheet_id'])

  #store.auth = project.task['auth']
  #store.trix_id = project.task.get('store', {}).get('sheet_id', project.task['sheet_id'])
  #store.load_id_map()

  video_format_dao = VideoFormatDAO(project.task['auth'],
                                    project.task['dcm_profile_id'],
                                    project.task['is_admin'])
  landing_page_dao = LandingPageDAO(project.task['auth'],
                                    project.task['dcm_profile_id'],
                                    project.task['is_admin'])
  placement_group_dao = PlacementGroupDAO(project.task['auth'],
                                          project.task['dcm_profile_id'],
                                          project.task['is_admin'])
  campaign_dao = CampaignDAO(project.task['auth'],
                             project.task['dcm_profile_id'],
                             project.task['is_admin'])
  creative_association_dao = CreativeAssociationDAO(
      project.task['auth'], project.task['dcm_profile_id'],
      project.task['is_admin'])
  creative_dao = CreativeDAO(project.task['auth'],
                             project.task['dcm_profile_id'],
                             project.task['is_admin'])
  placement_dao = PlacementDAO(project.task['auth'],
                               project.task['dcm_profile_id'],
                               project.task['is_admin'])
  creative_asset_dao = CreativeAssetDAO(project.task['auth'],
                                        project.task['dcm_profile_id'],
                                        project.task['is_admin'], project.id)
  ad_dao = AdDAO(project.task['auth'], project.task['dcm_profile_id'],
                 project.task['is_admin'])
  event_tag_dao = EventTagDAO(project.task['auth'],
                              project.task['dcm_profile_id'],
                              project.task['is_admin'])
  dynamic_targeting_key_dao = DynamicTargetingKeyDAO(
      project.task['auth'], project.task['dcm_profile_id'],
      project.task['is_admin'])


def assets():
  """Processes assets."""
  process_feed('creative_asset_feed', creative_asset_dao,
               FieldMap.CREATIVE_ASSET_FILE_NAME, 'Uploading creative asset')


def landing_pages():
  """Processes landing pages."""
  process_feed('landing_page_feed', landing_page_dao,
               FieldMap.CAMPAIGN_LANDING_PAGE_NAME, 'Processing landing page')


def placement_groups():
  """Processes placement group."""
  process_feed('placement_group_feed', placement_group_dao,
               FieldMap.PLACEMENT_GROUP_NAME, 'Processing placement group')


def campaigns():
  """Processes campaigns."""
  process_feed('campaign_feed', campaign_dao, FieldMap.CAMPAIGN_NAME,
               'Processing campaign')


def event_tags():
  """Processes event tags."""
  process_feed('event_tag_feed', event_tag_dao, FieldMap.EVENT_TAG_NAME,
               'Processing event tag')


def placements():
  """Processes placements."""
  placement_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'placement_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  pricing_schedule_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'placement_pricing_schedule_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  transcode_configs_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'transcode_configs_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  placement_dao.map_placement_transcode_configs(placement_feed.feed,
                                                transcode_configs_feed.feed,
                                                pricing_schedule_feed.feed)

  execute_feed(placement_feed, placement_dao, FieldMap.PLACEMENT_NAME,
               'Processing placement')

  pricing_schedule_feed.update()


def creatives():
  """Processes creatives."""
  creative_asset_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'creative_asset_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  creative_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'creative_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  third_party_url_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'third_party_url_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  creative_association_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'creative_asset_association_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  click_tag_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'click_tag_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  creative_dao.map_creative_third_party_url_feeds(creative_feed.feed,
                                                  third_party_url_feed.feed)

  creative_dao.map_creative_click_tag_feeds(creative_feed.feed,
                                            click_tag_feed.feed)

  third_party_url_feed.update()

  creative_dao.map_creative_and_association_feeds(
      creative_feed.feed, creative_association_feed.feed)

  creative_dao.map_assets_feed(creative_asset_feed)

  execute_feed(creative_feed, creative_dao, FieldMap.CREATIVE_NAME,
               'Processing creative')

  execute_feed(creative_association_feed, creative_association_dao,
               FieldMap.CREATIVE_ID, 'Associating with campaign, creative id')

  creative_association_feed.update()

  click_tag_feed.update()


def ads():
  """Processes ads."""
  placement_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'placement_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))
  event_tag_profile_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'event_tag_profile_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))
  ad_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'ad_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))
  ad_creative_assignment_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'ad_creative_assignment_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  ad_placement_assignment_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'ad_placement_assignment_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))
  ad_event_tag_assignment_feed = Feed(
      project.task['auth'],
      project.task['sheet_id'],
      'event_tag_ad_assignment_feed',
      spreadsheet=spreadsheet,
      timezone=project.task.get('timezone', None))

  ad_dao.map_feeds(ad_feed.feed, ad_creative_assignment_feed.feed,
                   ad_placement_assignment_feed.feed,
                   ad_event_tag_assignment_feed.feed, placement_feed.feed,
                   event_tag_profile_feed.feed)
  execute_feed(ad_feed, ad_dao, FieldMap.AD_ID, 'Processing Ad')

  ad_creative_assignment_feed.update()
  ad_placement_assignment_feed.update()
  ad_event_tag_assignment_feed.update()
  event_tag_profile_feed.update()


def dynamic_targeting_keys():
  """Processes dynamic targeting keys."""
  process_feed('dynamic_targeting_key_feed', dynamic_targeting_key_dao,
               FieldMap.DYNAMIC_TARGETING_KEY_NAME,
               'Processing dynamic targeting key')


@project.from_parameters
def traffic():
  """Main function of Bulkdozer, performs the Bulkdozer job"""
  global clean_run
  if project.verbose:
    print('traffic')

  try:
    setup()

    logger.clear()
    logger.log('Bulkdozer traffic job starting')
    logger.flush()

    init_daos()
    assets()
    landing_pages()
    campaigns()
    event_tags()
    placement_groups()
    placements()
    creatives()
    ads()
    dynamic_targeting_keys()

    #if clean_run:
    #  store.clear()

  except Exception as error:
    stack = traceback.format_exc()
    print(stack)

    logger.log(str(error))

  finally:
    logger.log('Bulkdozer traffic job ended')
    logger.flush()
    #store.save_id_map()

  if clean_run:
    print('Done: Clean run.')
  else:
    raise Exception(
        'Done: Errors happened with some of the assets, check your sheet log.')


@project.from_parameters
def _traffic():
  """For development purposes when debugging a specific entity, this function is handy to run just that entity."""
  setup()
  init_daos()
  dynamic_targeting_keys()


if __name__ == '__main__':
  """Main entry point of Bulkdozer."""
  traffic()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################


def Placement(name,
              compatibility,
              size_width,
              size_height,
              paymentSource,
              pricingSchedule_startDate,
              pricingSchedule_endDate,
              campaignId,
              placementStrategyId,
              contentCategoryId,
              siteId=None,
              directorySiteId=None):
  """ Default minial placement definition.

  Args:
    See docs:
      https://developers.google.com/doubleclick-advertisers/v3.2/placements#resource

  Returns:
    Placement JSON.

  """

  body = {
      'kind': 'dfareporting#placement',
      'campaignId': campaignId,
      'name': name,
      'siteId': siteId,
      'directorySiteId': directorySiteId,
      'paymentSource': paymentSource,
      'compatibility': compatibility,
      'size': {
          'width': size_width,
          'height': size_height
      },
      'archived': False,
      'pricingSchedule': {
          'testingStartDate': pricingSchedule_startDate,
          'startDate': pricingSchedule_startDate,
          'endDate': pricingSchedule_endDate,
          'pricingType': 'PRICING_TYPE_CPM',
          'capCostOption': 'CAP_COST_CUMULATIVE',
          'disregardOverdelivery': False,
          'flighted': False,
          'pricingPeriods': [{
              'startDate': pricingSchedule_startDate,
              'endDate': pricingSchedule_endDate,
              'units': 0,
              'rateOrCostNanos': 0,
              'pricingComment': ''
          }],
          'floodlightActivityId': None
      },
      'placementGroupId': None,
      'primary': False,
      'tagSetting': {
          'additionalKeyValues': None,
          'includeClickTracking': True,
          'includeClickThroughUrls': True,
          'keywordOption': 'IGNORE'
      },
      'tagFormats': [
          'PLACEMENT_TAG_STANDARD',
          'PLACEMENT_TAG_IFRAME_JAVASCRIPT',
          'PLACEMENT_TAG_IFRAME_ILAYER',
          'PLACEMENT_TAG_INTERNAL_REDIRECT',
          'PLACEMENT_TAG_JAVASCRIPT',
          'PLACEMENT_TAG_INTERSTITIAL_IFRAME_JAVASCRIPT',
          'PLACEMENT_TAG_INTERSTITIAL_INTERNAL_REDIRECT',
          'PLACEMENT_TAG_INTERSTITIAL_JAVASCRIPT',
          'PLACEMENT_TAG_CLICK_COMMANDS',
          'PLACEMENT_TAG_INSTREAM_VIDEO_PREFETCH',
          'PLACEMENT_TAG_INSTREAM_VIDEO_PREFETCH_VAST_3',
          'PLACEMENT_TAG_INSTREAM_VIDEO_PREFETCH_VAST_4',
          'PLACEMENT_TAG_TRACKING',
          'PLACEMENT_TAG_TRACKING_IFRAME',
          'PLACEMENT_TAG_TRACKING_JAVASCRIPT',
      ],
      'contentCategoryId': contentCategoryId,
      'placementStrategyId': placementStrategyId,
      'lookbackConfiguration': {
          'clickDuration': 14,
          'postImpressionActivitiesDuration': 14
      },
      'comment': '',
      'status': 'DRAFT',
      'sslRequired': False,
      'externalId': '',
      'videoSettings': {
          'kind': 'dfareporting#videoSettings',
          'companionSettings': {
              'kind': 'dfareporting#companionSetting',
              'companionsDisabled': False,
              'enabledSizes': [],
              'imageOnly': False
          },
          'transcodeSettings': {
              'kind': 'dfareporting#transcodeSetting',
              'enabledVideoFormats': []
          },
          'skippableSettings': {
              'kind': 'dfareporting#skippableSetting',
              'skippable': True,
              'skipOffset': {
                  'offsetSeconds': None,
                  'offsetPercentage': 50
              },
              'progressOffset': {
                  'offsetSeconds': None,
                  'offsetPercentage': 50
              }
          },
          'orientation': 'ANY'
      },
      'videoActiveViewOptOut': False,
      'vpaidAdapterChoice': 'BOTH',
      'adBlockingOptOut': False
  }

  # filter out conflicting fields
  if compatibility != 'IN_STREAM_VIDEO':
    del body['videoSettings']

  return body



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.patch import patch_clear
from starthinker.task.dv_editor.patch import patch_log
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def insertion_order_clear():
  table_create(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "DV_InsertionOrders",
    Discovery_To_BigQuery(
      "displayvideo",
      "v1").method_schema("advertisers.insertionOrders.list"
    )
  )

  sheets_clear(
    project.task["auth_sheets"],
    project.task["sheet"],
    "Insertion Orders",
    "A2:Z"
  )


def insertion_order_load():

  # load multiple partners from user defined sheet
  def insertion_order_load_multiple():
    rows = get_rows(
      project.task["auth_sheets"],
      { "sheets": {
        "sheet": project.task["sheet"],
        "tab": "Advertisers",
        "range": "A2:A"
      }}
    )

    for row in rows:
      yield from API_DV360(
        project.task["auth_dv"],
        iterate=True
      ).advertisers().insertionOrders().list(
        advertiserId=lookup_id(row[0])
      ).execute()

  # write insertion orders to database and sheet
  put_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table": "DV_InsertionOrders",
      "schema": Discovery_To_BigQuery(
        "displayvideo",
        "v1"
      ).method_schema("advertisers.insertionOrders.list"),
      "format": "JSON"
    }},
    insertion_order_load_multiple()
  )

  # write insertion orders to sheet
  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "query": """SELECT
        CONCAT(P.displayName, ' - ', P.partnerId),
        CONCAT(A.displayName, ' - ', A.advertiserId),
        CONCAT(C.displayName, ' - ', C.campaignId),
        CONCAT(I.displayName, ' - ', I.insertionOrderId),
        'PATCH',
        I.entityStatus,
        I.displayName,
        I.displayName,
        I.budget.budgetUnit,
        I.budget.budgetUnit,
        I.budget.automationType,
        I.budget.automationType,
        I.performanceGoal.performanceGoalType,
        I.performanceGoal.performanceGoalType,
        I.performanceGoal.performanceGoalAmountMicros / 1000000,
        I.performanceGoal.performanceGoalAmountMicros / 1000000,
        I.performanceGoal.performanceGoalPercentageMicros / 1000000,
        I.performanceGoal.performanceGoalPercentageMicros / 1000000,
        I.performanceGoal.performanceGoalString,
        I.performanceGoal.performanceGoalString
        FROM `{dataset}.DV_InsertionOrders` AS I
        LEFT JOIN `{dataset}.DV_Campaigns` AS C
        ON I.campaignId=C.campaignId
        LEFT JOIN `{dataset}.DV_Advertisers` AS A
        ON I.advertiserId=A.advertiserId
        LEFT JOIN `{dataset}.DV_Partners` AS P
        ON A.partnerId=P.partnerId
      """.format(**project.task),
      "legacy": False
    }}
  )

  put_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Insertion Orders",
      "range": "A2"
    }},
    rows
  )


def insertion_order_audit():

  # Move Insertion Order To BigQuery
  rows = get_rows(
    project.task["auth_sheets"], {
      "sheets": {
        "sheet": project.task["sheet"],
        "tab": "Insertion Orders",
        "range": "A2:Z"
      }
    }
  )

  put_rows(
    project.task["auth_bigquery"], {
      "bigquery": {
        "dataset": project.task["dataset"],
        "table": "SHEET_InsertionOrders",
        "schema": [
          { "name": "Partner", "type": "STRING" },
          { "name": "Advertiser", "type": "STRING" },
          { "name": "Campaign", "type": "STRING" },
          { "name": "Insertion_Order", "type": "STRING" },
          { "name": "Action", "type": "STRING" },
          { "name": "Status", "type": "STRING" },
          { "name": "Name", "type": "STRING" },
          { "name": "Name_Edit", "type": "STRING" },
          { "name": "Budget_Unit", "type": "STRING" },
          { "name": "Budget_Unit_Edit", "type": "STRING" },
          { "name": "Budget_Automation", "type": "STRING" },
          { "name": "Budget_Automation_Edit", "type": "STRING" },
          { "name": "Performance_Goal_Type", "type": "STRING" },
          { "name": "Performance_Goal_Type_Edit", "type": "STRING" },
          { "name": "Performance_Goal_Amount", "type": "FLOAT" },
          { "name": "Performance_Goal_Amount_Edit", "type": "FLOAT" },
          { "name": "Performance_Goal_Percent", "type": "FLOAT" },
          { "name": "Performance_Goal_Percent_Edit", "type": "FLOAT" },
          { "name": "Performance_Goal_String", "type": "STRING" },
          { "name": "Performance_Goal_String_Edit", "type": "STRING" },
        ],
        "format": "CSV"
      }
    },
    rows
  )

  # Create Insert View
  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "INSERT_InsertionOrders",
    """SELECT
      REGEXP_EXTRACT(S_IO.Advertiser, r' - (\d+)$') AS advertiserId,
      REGEXP_EXTRACT(S_IO.Campaign, r' - (\d+)$') AS campaignId,
      S_IO.Insertion_Order AS displayName,
      'ENTITY_STATUS_DRAFT' AS entityStatus,
      STRUCT(
        S_PC.Cost_Type_Edit As costType,
        S_PC.Fee_Type_Edit As feeType,
        S_PC.Invoice_Type_Edit AS invoiceType,
        S_PC.Fee_Amount_Edit AS feeAmount,
        S_PC.Fee_Percent_Edit * 1000 AS feePercentageMillis
      ) AS partnerCosts,
      STRUCT(
        S_P.Period_Edit As pacingPeriod,
        S_P.Type_Edit As pacingType,
        S_P.Daily_Budget_Edit AS dailyMaxMicros,
        S_P.Daily_Impressions_Edit AS dailyMaxImpressions
      ) AS pacing,
      STRUCT(
        S_FC.Unlimited_Edit AS unlimited,
        S_FC.Time_Unit_Edit AS timeUnit,
        S_FC.Time_Count_Edit AS timeUnitCount,
        S_FC.Max_impressions_Edit AS maxImpressions
      ) AS frequencyCap,
      STRUCT(
        S_ID.Integration_Code_Edit As integrationCode,
        S_ID.Details_Edit As details
      ) AS integrationDetails,
      STRUCT(
        S_IO.Performance_Goal_Type_Edit AS performanceGoalType,
        S_IO.Performance_Goal_Amount_Edit * 100000 AS performanceGoalAmountMicros,
        S_IO.Performance_Goal_Percent_Edit * 100000 AS performanceGoalPercentageMicros,
        S_IO.Performance_Goal_String_Edit AS performanceGoalString
      ) AS performanceGoal,
      STRUCT(
        S_IO.Budget_Unit_Edit AS budgetUnit,
        S_IO.Budget_Automation_Edit AS automationType,
        (SELECT ARRAY(
          SELECT
            STRUCT(
             S_S.Budget_Edit * 100000 AS budgetAmountMicros,
             S_S.Description_Edit AS description,
             STRUCT (
               STRUCT (
                 EXTRACT(YEAR FROM CAST(S_S.Start_Date_Edit AS Date)) AS year,
                 EXTRACT(MONTH FROM CAST(S_S.Start_Date_Edit AS DATE)) AS month,
                 EXTRACT(DAY FROM CAST(S_S.Start_Date_Edit AS DATE)) AS day
               ) AS startDate,
               STRUCT (
                 EXTRACT(YEAR FROM CAST(S_S.End_Date_Edit AS Date)) AS year,
                 EXTRACT(MONTH FROM CAST(S_S.End_Date_Edit AS DATE)) AS month,
                 EXTRACT(DAY FROM CAST(S_S.End_Date_Edit AS DATE)) AS day
             ) AS endDate
           ) AS dateRange
          ) AS budgetSegments
          FROM `{dataset}.SHEET_Segments` AS  S_S
          WHERE S_IO.Insertion_Order=S_S.Insertion_Order
        )) AS budgetSegments
      ) AS budget,
      STRUCT(
        IF(S_BS.Fixed_Bid_Edit IS NOT NULL,
          STRUCT(
            S_BS.Fixed_Bid_Edit * 100000 AS bidAmountMicros
          ),
          NULL
        ) AS fixedBid,
        IF(S_BS.Auto_Bid_Goal_Edit IS NOT NULL,
          STRUCT(
            S_BS.Auto_Bid_Goal_Edit AS performanceGoalType,
            S_BS.Auto_Bid_Amount_Edit * 100000 AS maxAverageCpmBidAmountMicros,
            S_BS.Auto_Bid_Algorithm_Edit AS customBiddingAlgorithmId
          ),
          NULL
        ) AS maximizeSpendAutoBid,
        IF(S_BS.Performance_Goal_Type_Edit IS NOT NULL,
          STRUCT(
            S_BS.Performance_Goal_Type_Edit AS performanceGoalType,
            S_BS.Performance_Goal_Amount_Edit * 100000 AS performanceGoalAmountMicros,
            S_BS.Performance_Goal_Average_CPM_Bid_Edit * 100000 AS maxAverageCpmBidAmountMicros,
            S_BS.Performance_Goal_Algorithm_Edit AS customBiddingAlgorithmId
          ),
          NULL
        ) AS performanceGoalAutoBid
      )
      AS bidStrategy
      FROM `{dataset}.SHEET_InsertionOrders` As S_IO
      LEFT JOIN `{dataset}.SHEET_Segments` As S_S ON S_IO.Insertion_Order=S_S.Insertion_Order
      LEFT JOIN `{dataset}.SHEET_PartnerCosts` As S_PC ON S_IO.Insertion_Order=S_PC.Insertion_Order
      LEFT JOIN `{dataset}.SHEET_Pacing` As S_P ON S_IO.Insertion_Order=S_P.Insertion_Order
      LEFT JOIN `{dataset}.SHEET_FrequencyCaps` As S_FC ON S_IO.Insertion_Order=S_FC.Insertion_Order
      LEFT JOIN `{dataset}.SHEET_IntegrationDetails` As S_ID ON S_IO.Insertion_Order=S_ID.Insertion_Order
      LEFT JOIN `{dataset}.SHEET_BidStrategy` As S_BS ON S_IO.Insertion_Order=S_BS.Insertion_Order
      LEFT JOIN `{dataset}.DV_InsertionOrders` As DV_IO ON S_IO.Insertion_Order=DV_IO.displayName
      WHERE S_IO.Action="INSERT"
      AND DV_IO IS NULL
    """.format(**project.task),
    legacy=False
  )

  # Create Audit View And Write To Sheets
  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "AUDIT_InsertionOrders",
    """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Insertion Order' AS Operation,
            CASE
              WHEN Name_Edit IS NULL THEN 'Missing Name.'
              WHEN Budget_Unit_Edit IS NULL THEN 'Missing Budget Unit.'
              WHEN Budget_Automation_Edit IS NULL THEN 'Missing Budget Automation.'
              WHEN Performance_Goal_Type_Edit IS NULL THEN 'Missing Goal Type.'
              WHEN Performance_Goal_Amount_Edit IS NULL
              AND Performance_Goal_Percent_Edit IS NULL
              AND Performance_Goal_String_Edit IS NULL THEN 'Missing Goal Amount / Percent / String.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_InsertionOrders`
        )
        WHERE
          Error IS NOT NULL
      ),
      /* Check duplicate inserts */
      DUPLICATE_ERRORS AS (
        SELECT
          'Insertion_Order' AS Operation,
          'Duplicate Insertion Order name, insert will be ignored.' AS Error,
          'WARNING' AS Severity,
          COALESCE(S_IO.Insertion_Order, 'BLANK') AS Id
        FROM `{dataset}.SHEET_InsertionOrders` As S_IO
        LEFT JOIN `{dataset}.DV_InsertionOrders` AS DV_IO ON S_IO.Insertion_Order=DV_IO.displayName
        WHERE S_IO.Action="INSERT"
        AND DV_IO IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      UNION ALL
      SELECT * FROM DUPLICATE_ERRORS
      ;
    """.format(**project.task),
    legacy=False
  )

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_InsertionOrders",
    """SELECT *
      FROM `{dataset}.SHEET_InsertionOrders`
      WHERE Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_InsertionOrders` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def insertion_order_patch(commit=False):

  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_InsertionOrders",
    }},
    as_object=True
  )

  for row in rows:
    if row['Action'] == "DELETE":
      patches.append({
        "operation": "Insertion Orders",
        "action": "DELETE",
        "partner": row['Partner'],
        "advertiser": row['Advertiser'],
        "campaign": row['Campaign'],
        "insertion_order": row['Insertion_Order'],
        "parameters": {
          "advertiserId": lookup_id(row['Advertiser']),
          "insertionOrderId": lookup_id(row['Insertion_Order'])
        }
      })

    elif row['Action'] == "PATCH":
      insertion_order = {}

      if row['Name'] != row['Name_Edit']:
        insertion_order["displayName"] = row['Name_Edit']

      if row['Budget_Unit'] != row['Budget_Unit_Edit']:
        insertion_order.setdefault("budget", {})
        insertion_order["budget"]["budgetUnit"] = row['Budget_Unit_Edit']

      if row['Budget_Automation'] != row['Budget_Automation_Edit']:
        insertion_order.setdefault("budget", {})
        insertion_order["budget"]["automationType"] = row['Budget_Automation_Edit']

      if row['Performance_Goal_Type'] != row['Performance_Goal_Type_Edit']:
        insertion_order.setdefault("performanceGoal", {})
        insertion_order["performanceGoal"]["performanceGoalType"] = row['Performance_Goal_Type_Edit']

      if row['Performance_Goal_Amount'] != row['Performance_Goal_Amount_Edit']:
        insertion_order.setdefault("performanceGoal", {})
        insertion_order["performanceGoal"]["performanceGoalAmountMicros"] = int(
          float(row['Performance_Goal_Amount_Edit']) * 1000000
        )

      if row['Performance_Goal_Percent'] != row['Performance_Goal_Percent_Edit']:
        insertion_order.setdefault("performanceGoal", {})
        insertion_order["performanceGoal"]["performanceGoalPercentageMicros"] = int(
          float(row['Performance_Goal_Percent_Edit']) * 1000000
        )

      if row['Performance_Goal_String'] != row['Performance_Goal_String_Edit']:
        insertion_order.setdefault("performanceGoal", {})
        insertion_order["performanceGoal"]["performanceGoalString"] = row['Performance_Goal_String_Edit']

      if insertion_order:
        patches.append({
          "operation": "Insertion Orders",
          "action": "PATCH",
          "partner": row['Partner'],
          "advertiser": row['Advertiser'],
          "campaign": row['Campaign'],
          "insertion_order": row['Insertion_Order'],
          "parameters": {
            "advertiserId": lookup_id(row['Advertiser']),
            "insertionOrderId": lookup_id(row['Insertion_Order']),
            "body": insertion_order
          }
        })

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
  else:
    patch_preview(patches)


def insertion_order_insert(commit=False):
  inserts = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"INSERT_InsertionOrders",
    }},
    as_object=True
  )

  for row in rows:
    inserts.append({
      "operation": "Insertion Orders",
      "action": "INSERT",
      "partner": None,
      "advertiser": row['advertiserId'],
      "campaign": row['campaignId'],
      "insertion_order": row['displayName'],
      "parameters": {
        "advertiserId": row['advertiserId'],
        "body":row
      }
    })

  if commit:
    insertion_order_commit(inserts)
  else:
    patch_preview(inserts)


def insertion_order_commit(patches):
  for patch in patches:
    if not patch.get("insertion_order"):
      continue
    print("API INSERTION ORDER:", patch["action"], patch["insertion_order"])
    try:
      if patch["action"] == "DELETE":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().insertionOrders().delete(
          **patch["parameters"]
        ).execute()
        patch["success"] = response
      elif patch["action"] == "PATCH":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().insertionOrders().patch(
          **patch["parameters"]
        ).execute()
        patch["success"] = response["insertionOrderId"]
      elif patch["action"] == "INSERT":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().insertionOrders().create(
          **patch["parameters"]
        ).execute()
        patch["success"] = response["insertionOrderId"]
    except Exception as e:
      patch["error"] = str(e)
    finally:
      patch_log(patch)
  patch_log()



"""One Platform python media compatibility tests."""

import random
import string
import StringIO

import google3
import apitools.base.py as apitools_base

from google3.google.api.hexa import one_platform_env
from google3.pyglib import flags
from google3.testing.integration.hexa import testcase

from google3.apiserving.tools.testing.media import media_v1_client as media_client
from google3.apiserving.tools.testing.media import media_v1_messages as media_messages

FLAGS = flags.FLAGS
_CLIENT = None


def _GenerateData(length, alphabet=None):
  if alphabet is None:
    alphabet = string.hexdigits[:16]
  return ''.join(random.choice(alphabet) for _ in xrange(length))


def _GetClient(api_endpoint):
  global _CLIENT
  if _CLIENT is None:
    _CLIENT = media_client.MediaV1(
        api_endpoint,
        get_credentials=False,
        additional_http_headers={'X-GFE-SSL': 'yes'})
  return _CLIENT


class PythonMediaTest(one_platform_env.OnePlatformEnvComponentsMixin,
                      testcase.HexaTestCase):

  def setUp(self):
    api_endpoint = self.stash.apiComponent.GetProperties()['api_url']
    self.client = _GetClient(api_endpoint)

  def testMc1SimpleUpload(self):
    data = _GenerateData(32 * 1024)
    resource_name = 'objects/non_resumable'

    upload_stream = StringIO.StringIO(data)
    upload = apitools_base.Upload.FromStream(upload_stream, 'text/plain')
    upload.strategy = 'simple'
    upload_request = media_messages.Media(resourceName=resource_name)
    upload_response = self.client.media.Upload(upload_request, upload=upload)
    self.assertIsNotNone(upload_response)

    # Download process

    resource_name = 'sdfdownloadtask/media/23362726'
    download_stream = StringIO.StringIO()
    download = apitools_base.Download.FromStream(download_stream)
    download_request = media_messages.MediaMediaDownloadRequest(
        resourceName=resource_name)

    client = media_client.MediaV1(
        api_endpoint,
        get_credentials=False,
        additional_http_headers={'X-GFE-SSL': 'yes'})

    client.media.Download(download_request, download=download)
    download_stream.seek(0)
    self.assertEqual(data, download_stream.getvalue())

  def testMc2ResumableUpload(self):
    """Resumable upload not yet implemented in ESF."""


if __name__ == '__main__':
  testcase.main()



import time
from googleapiclient.http import MediaIoBaseUpload, MediaIoBaseDownload
from starthinker.util.data import get_rows
from starthinker.util.auth import get_credentials
from starthinker.util.storage import media_download
from starthinker.util.csv import column_header_sanitize

from googleapiclient import discovery

from starthinker.util.storage import object_put

from starthinker.util.bigquery import io_to_table, table_create

from starthinker.task.sdf.schema.Lookup import SDF_Field_Lookup

from starthinker.util.google_api import API_DV360

import io
import zipfile
from io import BytesIO

import apitools.base.py as apitools_base

LINE_ITEM_SCHEMA = [{
    'name': 'Line_Item_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Io_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Subtype',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Timestamp',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Start_Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'End_Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Budget_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Budget_Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Pacing',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Pacing_Rate',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Pacing_Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Frequency_Enabled',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Frequency_Exposures',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Frequency_Period',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Frequency_Amount',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Trueview_View_Frequency_Enabled',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Trueview_View_Frequency_Exposures',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Trueview_View_Frequency_Period',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_Revenue_Model',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_Revenue_Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Counting_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Counting_Pct',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Floodlight_Activity_Ids',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Fees',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Integration_Code',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Details',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Strategy_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Strategy_Value',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Strategy_Unit',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Strategy_Do_Not_Exceed',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Apply_Floor_Price_For_Deals',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Assignments',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Geography_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Geography_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Geography_Regional_Location_List_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Geography_Regional_Location_List_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Language_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Language_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Device_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Device_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Digital_Content_Labels_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Brand_Safety_Sensitivity_Setting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Brand_Safety_Custom_Settings',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Third_Party_Verification_Services',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Third_Party_Verification_Labels',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Channel_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Channel_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'App_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'App_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'App_Collection_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'App_Collection_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Category_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Category_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Keyword_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Keyword_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Keyword_List_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeting_Similar_Audiences',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Affinity_and_In_Market_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Affinity_and_In_Market_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_List_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Inventory_Source_Targeting_Authorized_Seller_Options',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Inventory_Source_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Inventory_Source_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Inventory_Source_Targeting_Target_New_Exchanges',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Daypart_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Daypart_Targeting_Time_Zone',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Environment_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Viewability_Targeting_Active_View',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Position_Targeting_On_Screen',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Position_Targeting_Display_Position_In_Content',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Position_Targeting_Video_Position_In_Content',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Position_Targeting_Audio_Position_In_Content',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Size_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Gender',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Age',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Household_Income',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Parental_Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Connection_Speed_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Carrier_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Carrier_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Multipliers',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Video_Ad_Formats',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Bid_Strategy_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Bid_Strategy_Value',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Mobile_Bid_Adjustment_Option',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Mobile_Bid_Adjustment_Percentage',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Desktop_Bid_Adjustment_Option',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Desktop_Bid_Adjustment_Percentage',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Tablet_Bid_Adjustment_Option',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Tablet_Bid_Adjustment_Percentage',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Connected_TV_Bid_Adjustment_Option',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Connected_TV_Bid_Adjustment_Percentage',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Category_Exclusions_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Content_Filter',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Inventory_Source_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Third_Party_Viewability_Vendor',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'TrueView_Third_Party_Brand_Safety_Vendor',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]

AD_GROUP_SCHEMA = [{
    'name': 'Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Line_Item_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Ad_Format',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid_Cost',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Popular_Videos_Bid_Adjustment',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Keyword_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Keyword_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Category_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Category_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_Channels_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_Channels_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_Videos_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_Videos_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_URLs_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_YouTube_URLs_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_Apps_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_Apps_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_App_Collections_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Targeting_App_Collections_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Gender',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Age',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Household_Income',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Demographic_Targeting_Parental_Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Affinity_and_In_Market_Targeting_Include',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Affinity_and_In_Market_Targeting_Exclude',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_List_Targeting',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]


def sdf_read(project):
  # Read Filter Ids
  filter_ids = list(
      get_rows(project.task['auth'], project.task['sdf']['read']['filter_ids']))

  # todo handle inventory sources
  body = {
      'version': project.task['sdf']['version'],
      'partnerId': project.task['sdf']['partner_id'],
      'parentEntityFilter': {
          'fileType': project.task['sdf']['file_types'],
          'filterType': project.task['sdf']['filter_type'],
          'filterIds': filter_ids
      },
      'idFilter': None
  }

  operation = API_DV360('user').sdfdownloadtasks().create(body=body).execute()

  print('operation====================')
  print(operation)

  if operation or 'name' not in operation:
    request = API_DV360('user').sdfdownloadtasks().operations().get(
        name=operation['name'])
    # This is the eng recommended way of getting the operation
    while True:
      response = request.execute()
      if 'done' in response and response['done']:
        break
      time.sleep(30)
  else:
    print('error')
    #todo error checking if operation is non null

  print('response====================')
  print(response)

  request = API_DV360('user').media().download_media(resourceName=resource_name)

  data = media_download(request, CHUNK_SIZE, encoding=None)


def dt_move_large(data, table_name, jobs):
  if project.verbose:
    print('DT TO TABLE LARGE', dt_partition)

  delimiter = '\n'
  disposition = 'WRITE_TRUNCATE'

  # decompression handler for gzip ( must be outside of chunks as it keeps track of stream across multiple calls )
  gz_handler = zlib.decompressobj(32 + zlib.MAX_WBITS)

  # sliding view of data flowing out of decompression, used to buffer and delimit rows
  first_row = True
  view = ''

  # loop all chunks of file, decompress, and find row delimiter
  for data_gz in data:

    view += gz_handler.decompress(data_gz.read()).decode()

    if first_row:
      end = view.find(delimiter)
      schema = dt_schema(view[:end].split(','))
      view = view[(end + 1):]
      first_row = False

    end = view.rfind(delimiter)

    jobs.append(
        io_to_table(project.task['auth'], project.id,
                    project.task['to']['dataset'], table_name,
                    StringIO(view[:end]), 'CSV', schema, 0, disposition, False))
    disposition = 'WRITE_APPEND'
    view = view[min(end + 1, len(view)):]


def sdf_schema(header):
  schema = []
  for h in header:
    h = column_header_sanitize(h)
    schema.append({
        'name': h,
        'type': DT_Field_Lookup.get(h, 'STRING'),
        'mode': 'NULLABLE'
    })
  return schema

  # with zipfile.ZipFile(data) as d:
  #   print(d)
  #   print(d.infolist())
  #   print(d.namelist())
  #   file_names = d.namelist()
  #   for file_name  in file_names:
  #     print('file name======================================')
  #     print(file_name)
  #     if 'Skipped' in file_name:
  #       continue
  #     with d.open(file_name) as file:

  #       wrapper = io.TextIOWrapper(file, encoding='utf-8')
  #       schema = None

  #       #todo terwilleger figure out the schema piece of this
  #       # remember that the header was actually in the document so could try that

  #       if file_name == 'SDF-AdGroups.csv':
  #         schema = AD_GROUP_SCHEMA
  #       elif file_name == 'SDF-LineItems.csv':
  #         schema = LINE_ITEM_SCHEMA

  #       table_name = 'SDF_' + file_name.split('.')[0].split('-')[1]
  #       disposition = 'WRITE_TRUNCATE'
  #       if project.task['sdf']['daily']:
  #         disposition = 'WRITE_APPEND'
  #         table_create('service', project.id, project.task['sdf']['out']['bigquery']['dataset'], table_name, is_time_partition=True)
  #         #todo terwilleger add something here that creates the time partitioned table if it doesn't exist

  #       io_to_table('service', project.id, project.task['sdf']['out']['bigquery']['dataset'], table_name, wrapper, schema=schema, skip_rows=1, disposition=disposition)


#todo flag for if delete temp tables


def sdf_schema(header):
  schema = []

  for h in header:
    schema.append({
        'name': column_header_sanitize(h),
        'type': SDF_Field_Lookup.get(h, 'STRING'),
        'mode': 'NULLABLE'
    })

  return schema


def download_media(auth, resource_name):
  data = BytesIO()

  request = API_DV360('user').media().download_media(resourceName=resource_name)

  downloader = MediaIoBaseDownload(data, request)

  download_finished = False
  while download_finished is False:
    _, download_finished = downloader.next_chunk()

  return data



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import time

from airflow import models
from airflow import utils as airflow_utils


class Hello(models.BaseOperator):

  @airflow_utils.apply_defaults
  def __init__(self, say='', sleep=0, **kwargs):
    self.say = say
    self.sleep = sleep
    super(Hello, self).__init__(**kwargs)

  def execute(self, context):
    print(self.say)
    if self.sleep:
      time.sleep(self.sleep)
